dict_of_parts = {
    "BTS7960": "2 x",
    "connector": "2 females",
    "white-terminal": "8 x",
    "nuts-bolts-spacers": "M3: ({})".format(
        " + ".join(
            [
                "8 x 25 mm spacer",
                "4 x 30 mm spacer",
                "4 x nut",
            ]
        )
    ),
    "solid-cable-1-15": "20 cm x (red + black/blue)",
}



import os

from blueness import module
from bluer_options import string
from bluer_options import host
from bluer_options.logger import crash_report
from bluer_options.timer import Timer
from bluer_objects import file
from bluer_objects import objects
from bluer_objects.storage import upload
from bluer_objects.graphics.signature import add_signature
from bluer_ai import VERSION as abcli_VERSION
from bluer_ai.modules import terraform
from bluer_objects.env import abcli_object_name

from bluer_sbc import NAME
from bluer_sbc import env
from bluer_sbc.host import signature
from bluer_sbc.session.functions import reply_to_bash
from bluer_sbc.algo.diff import Diff
from bluer_sbc.hardware import hardware
from bluer_sbc.imager import imager
from bluer_sbc.logger import logger


NAME = module.name(__file__, NAME)


class Session:
    def __init__(self):
        self.bash_keys = {
            "i": "exit",
            "o": "shutdown",
            "p": "reboot",
            "u": "update",
        }

        self.diff = Diff(env.BLUER_SBC_SESSION_IMAGER_DIFF)

        self.capture_requested = False

        self.frame = 0
        self.new_frame = False
        self.frame_image = terraform.poster(None)
        self.frame_filename = ""

        self.auto_upload = env.BLUER_SBC_SESSION_AUTO_UPLOAD

        self.messages = []

        self.model = None

        self.params = {"iteration": -1}

        self.state = {}

        self.timer = {}
        for name, period in {
            "imager": env.BLUER_SBC_SESSION_IMAGER_PERIOD,
            "messenger": env.BLUER_SBC_SESSION_MESSENGER_PERIOD,
            "reboot": env.BLUER_SBC_SESSION_REBOOT_PERIOD,
            "screen": env.BLUER_SBC_SESSION_SCREEN_PERIOD,
            "temperature": env.BLUER_SBC_SESSION_TEMPERATURE_PERIOD,
        }.items():
            self.add_timer(name, period)

    def add_timer(
        self,
        name: str,
        period: float,
    ):
        if name not in self.timer:
            self.timer[name] = Timer(period, name)
            logger.info(
                "{}: timer[{}]:{}".format(
                    NAME,
                    name,
                    string.pretty_frequency(1 / period),
                )
            )
            return True
        return False

    def check_imager(self):
        self.new_frame = False

        if not env.BLUER_SBC_SESSION_IMAGER_ENABLED:
            return
        if not self.capture_requested and not self.timer["imager"].tick():
            return
        self.capture_requested = False

        if env.BLUER_SBC_CAMERA_KEEP_OPEN:
            if imager.device is None:
                if not imager.open(log=True):
                    return

        success, image = imager.capture(  # pylint: disable=unexpected-keyword-arg
            open_before=not bool(env.BLUER_SBC_CAMERA_KEEP_OPEN),
            close_after=not bool(env.BLUER_SBC_CAMERA_KEEP_OPEN),
        )
        if not success:
            return

        hardware.pulse("data")

        if self.diff.same(image):
            return

        self.frame += 1

        image = add_signature(
            image,
            [" | ".join(objects.signature(self.frame))],
            [" | ".join(signature())],
        )

        filename = objects.path_of(
            object_name=abcli_object_name,
            filename=f"{self.frame:016d}.jpg",
        )
        if not file.save_image(filename, image):
            return

        self.new_frame = True
        self.frame_image = image
        self.frame_filename = filename

        if self.auto_upload:
            upload(
                object_name=abcli_object_name,
                filename=file.name_and_extension(self.frame_filename),
            )

    def check_keys(self):
        for key in hardware.key_buffer:
            if key in self.bash_keys:
                reply_to_bash(self.bash_keys[key])
                return False

        if " " in hardware.key_buffer:
            self.capture_requested = True

        hardware.key_buffer = []

        return None

    def check_seed(self):
        seed_filename = host.get_seed_filename()
        if not file.exists(seed_filename):
            return None

        success, content = file.load_json(file.set_extension(seed_filename, "json"))
        if not success:
            return None

        hardware.pulse("outputs")

        seed_version = content.get("version", "")
        if seed_version <= abcli_VERSION:
            return None

        logger.info(f"{NAME}: seed {seed_version} detected.")
        reply_to_bash("seed", [seed_filename])
        return False

    def check_timers(self):
        if self.timer["screen"].tick():
            hardware.update_screen(
                image=self.frame_image,
                session=self,
                header=self.signature(),
            )
        elif hardware.animated:
            hardware.animate()

        if self.timer["reboot"].tick("wait"):
            reply_to_bash("reboot")
            return False

        if self.timer["temperature"].tick():
            self.read_temperature()

        return None

    def close(self):
        hardware.release()

        if env.BLUER_SBC_CAMERA_KEEP_OPEN:
            imager.close(log=True)

    def process_message(self, message):
        if (
            env.BLUER_SBC_SESSION_OUTBOUND_QUEUE
            and message.subject in "bolt,frame".split(",")
            and not host.is_headless()
        ):
            logger.info(f"{NAME}: frame received: {message.as_string()}")
            self.new_frame, self.frame_image = file.load_image(message.filename)

        if message.subject == "capture":
            logger.info(f"{NAME}: capture message received.")
            self.capture_requested = True

        if message.subject in "reboot,shutdown".split(","):
            logger.info(f"{NAME}: {message.subject} message received.")
            reply_to_bash(message.subject)
            return False

        if message.subject == "update":
            try:
                if message.data["version"] > abcli_VERSION:
                    reply_to_bash("update")
                    return False
            except Exception as e:
                crash_report(e)

        return None

    # https://www.cyberciti.biz/faq/linux-find-out-raspberry-pi-gpu-and-arm-cpu-temperature-command/
    def read_temperature(self):
        if not host.is_rpi():
            return

        params = {}

        success, output = file.load_text("/sys/class/thermal/thermal_zone0/temp")
        if success:
            output = [thing for thing in output if thing]
            if output:
                try:
                    params["temperature.cpu"] = float(output[0]) / 1000
                except Exception as e:
                    crash_report(e)
                    return

        self.params.update(params)
        logger.info(
            "{}: {}".format(
                NAME,
                ", ".join(string.pretty_param(params)),
            )
        )

    def signature(self):
        return [
            " | ".join(objects.signature()),
            " | ".join(sorted([timer.signature() for timer in self.timer.values()])),
            " | ".join(
                (["*"] if self.new_frame else [])
                + (["^"] if self.auto_upload else [])
                + hardware.signature()
                + [
                    "diff: {:.03f} - {}".format(
                        self.diff.last_diff,
                        string.pretty_duration(
                            self.diff.last_same_period,
                            largest=True,
                            include_ms=True,
                            short=True,
                        ),
                    ),
                    string.pretty_shape_of_matrix(self.frame_image),
                ]
                + ([] if self.model is None else self.model.signature())
            ),
        ]

    @staticmethod
    def start():
        success = True
        logger.info(f"{NAME}: started ...")

        try:
            session = Session()

            while session.step():
                pass

            logger.info(f"{NAME}: stopped.")
        except KeyboardInterrupt:
            logger.info(f"{NAME}: Ctrl+C: stopped.")
            reply_to_bash("exit")
        except Exception as e:
            crash_report(e)
            success = False

        try:
            session.close()
        except Exception as e:
            crash_report(e)
            success = False

        return success

    def step(
        self,
        steps="all",
    ) -> bool:
        if steps == "all":
            steps = "imager,keys,messages,seed,switch,timers".split(",")

        self.params["iteration"] += 1

        hardware.pulse("loop", 0)

        for enabled, step_ in zip(
            [
                "keys" in steps,
                "timers" in steps,
                "seed" in steps,
                "imager" in steps,
            ],
            [
                self.check_keys,
                self.check_timers,
                self.check_seed,
                self.check_imager,
            ],
        ):
            if not enabled:
                continue
            output = step_()
            if output in [False, True]:
                return output

            hardware.clock()

        return True



from bluer_sbc.README.designs.cheshmak import marquee as cheshmak_marquee
from bluer_sbc.README.designs.blue_bracket import items as blue_bracket_items
from bluer_sbc.README.designs.swallow import marquee as swallow_marquee
from bluer_sbc.README.designs.swallow_head import marquee as swallow_head_marquee
from bluer_sbc.README.designs.bryce import marquee as bryce_marquee
from bluer_sbc.README.designs.nafha import marquee as nafha_marquee
from bluer_sbc.README.designs.shelter import marquee as shelter_marquee
from bluer_sbc.README.designs.adapter_bus import marquee as adapter_bus_marquee
from bluer_sbc.README.designs.battery_bus import marquee as battery_bus_marquee
from bluer_sbc.README.designs.ultrasonic_sensor_tester import (
    marquee as ultrasonic_sensor_tester_marquee,
)
from bluer_sbc.README.shortcuts import items as shortcuts_items

docs = [
    {
        "items": []
        + swallow_head_marquee
        + swallow_marquee
        + battery_bus_marquee
        + adapter_bus_marquee
        + ultrasonic_sensor_tester_marquee
        + bryce_marquee
        + cheshmak_marquee
        + nafha_marquee
        + shelter_marquee
        + blue_bracket_items,
        "path": "../..",
        "macros": {
            "shortcuts:::": shortcuts_items,
        },
    },
]



import random

from bluer_sbc import fullname
from bluer_sbc.logger import logger


class Hardware:
    def __init__(self):
        logger.info(f"{self.__class__.__name__}.init().")

        self.key_buffer = []
        self.animated = False

    def animate(self):
        if self.buffer is None:
            return self
        if not self.animated:
            return self

        y = random.randint(0, self.buffer.shape[0] - 1)
        x = random.randint(0, self.buffer.shape[1] - 1)

        self.buffer[y, x] = 255 - self.buffer[y, x]

        self.animated = False
        self.update_screen(self.buffer, None, [], [])
        self.animated = True

    def clock(self):
        return self

    def pressed(self, keys):
        output = bool([key for key in keys if key in self.key_buffer])

        self.key_buffer = [key for key in self.key_buffer if key not in keys]

        return output

    def pulse(self, pin=None, frequency=None):
        """
        pulse pin.
        :param pin: "data" / "incoming" / "loop" / "outputs"
        :param frequency: frequency
        :return: self
        """
        return self

    def release(self):
        logger.info(f"{self.__class__.__name__}.release()")

    def signature(self):
        return [
            fullname(),
            f"hardware:{self.__class__.__name__}",
        ]

    def update_screen(self, image, session, header):
        return self



import argparse
import time

from blueness import module

from bluer_sbc import NAME
from bluer_sbc.hardware.sparkfun_top_phat.classes import Sparkfun_Top_phat
from bluer_sbc.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    default="",
    help="validate_leds",
)
args = parser.parse_args()

hardware = Sparkfun_Top_phat()

success = False
if args.task == "validate_leds":
    logger.info("loop started (Ctrl+C to stop)")
    offset = 0
    # https://stackoverflow.com/a/18994932/10917551
    try:
        while True:
            for index in range(hardware.pixel_count):
                hardware.pixels[index] = tuple(
                    int(thing * hardware.intensity)
                    for thing in hardware.colormap[
                        (index + offset) % hardware.pixel_count
                    ][:3]
                )

            offset += 1

            hardware.pixels.show()
            time.sleep(0.1)
    except KeyboardInterrupt:
        logger.info("Ctrl+C, stopping.")
    finally:
        hardware.release()
    success = True
else:
    logger.error(f"-{NAME}: {args.task}: command not found.")

if not success:
    logger.error(f"-{NAME}: {args.task}: failed.")



from bluer_objects import README
from bluer_objects.README.items import ImageItems
from bluer_objects.README.consts import assets_url

assets2 = assets_url(
    suffix="shelter",
    volume=2,
)

image_template = assets2 + "/{}?raw=true"


marquee = README.Items(
    [
        {
            "name": "shelter",
            "marquee": f"{assets2}/20251006_181554.jpg",
            "url": "./bluer_sbc/docs/shelter.md",
        }
    ]
)

items = ImageItems(
    {image_template.format(f"{index+1:02}.png"): "" for index in range(4)}
) + ImageItems(
    {
        f"{assets2}/20251005_180841.jpg": "",
        f"{assets2}/20251006_181432.jpg": "",
        f"{assets2}/20251006_181509.jpg": "",
        f"{assets2}/20251006_181554.jpg": "",
        f"{assets2}/20251028_113245.jpg": "",
    }
)



from bluer_objects import README
from bluer_objects.README.items import ImageItems
from bluer_objects.README.consts import assets_url, designs_url

from bluer_sbc.designs.adapter_bus.parts import dict_of_parts as parts


assets2 = assets_url(
    suffix="adapter-bus",
    volume=2,
)

marquee = README.Items(
    [
        {
            "name": "adapter bus",
            "marquee": f"{assets2}/20251017_222911.jpg",
            "url": "./bluer_sbc/docs/adapter-bus.md",
        }
    ]
)

items = ImageItems(
    {
        designs_url(
            "adapter-bus/wiring.png?raw=true",
        ): designs_url(
            "adapter-bus/wiring.svg",
        ),
        **{
            f"{assets2}/{timestamp}.jpg": ""
            for timestamp in [
                "20251017_222911",
                "20251017_222929",
                "20251017_222938",
                "20251017_222943",
                "20251017_222949",
                "20251017_223017",
                "20251017_223034",
                "20251018_213244",
            ]
        },
    }
)



dict_of_parts = {
    "rpi": "",
    "XL4015": "",
    "470-mF": "",
    "Polyfuse": "optional",
    "TVS-diode": "",
    "resistor": "7 x 330-470 Ω + N x 2.2 kΩ + N x 3.3 kΩ",
    "LED": "green + red + yellow + 4 x blue",
    "rpi-camera": "",
    "PCB-single-14x9_5": "2 x",
    "PCB-double-9x7": "2 x",
    "pushbutton": "",
    "ultrasonic-sensor": "4 x",
    "connector": "1 female",
    "nuts-bolts-spacers": " + ".join(
        [
            "M2: ({})".format(
                " + ".join(
                    [
                        "2 x bolt",
                        "2 x 5 mm spacer",
                        "4 x nut",
                    ]
                )
            ),
            "M2.5: ({})".format(
                " + ".join(
                    [
                        "4 x bolt",
                        "8 x 10 mm spacer",
                        "4 x nut",
                    ]
                )
            ),
            "M3: ({})".format(
                " + ".join(
                    [
                        "1 x bolt",
                        "3 x 35 mm spacer",
                        "3 x 25 mm spacer",
                        "7 x 15 mm spacer",
                        "4 x 5 mm spacer",
                        "5 x nut",
                    ]
                )
            ),
        ]
    ),
    "plexiglass": "14 cm x 9.5 cm",
    "white-terminal": "2 x",
    "dupont-cables": "1 x 30 cm + 1 x 10 cm",
    "16-awg-wire": "40 cm x (red + black/blue)",
    "solid-cable-1-15": "10 cm x (red + black/blue)",
    "strong-thread": "1 m",
    "pin-headers": "1 x (female, 2 x 40) -> 2 x 20 + 2 x (male, 1 x 40) -> 4 x 6 + 4 x 2 + 2 x 20",
}



from typing import List

from bluer_options.terminal import show_usage


def help_info(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "grove",
            "info",
        ],
        "show grove info.",
        mono=mono,
    )


def help_validate(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "adc | button"

    return show_usage(
        [
            "grove",
            "validate",
            f"[{options}]",
        ],
        "validate grove.",
        mono=mono,
    )


def help_validate_oled_128x64(
    tokens: List[str],
    mono: bool,
) -> str:
    options = "animate | buttons | image | shapes | stats"

    return show_usage(
        [
            "grove",
            "validate",
            "oled_128x64",
            f"[{options}]",
        ],
        "validate grove oled_128x64.",
        mono=mono,
    )


help_functions = {
    "info": help_info,
    "validate": help_validate,
    "validate_oled_128x64": help_validate_oled_128x64,
}



import time

from blueness import module
from bluer_options import string
from bluer_options import host

from bluer_sbc import NAME
from bluer_sbc.hardware.hat.abstract import Abstract_Hat
from bluer_sbc.logger import logger

NAME = module.name(__file__, NAME)


class Prototype_Hat(Abstract_Hat):
    def __init__(self):
        super().__init__()

        self.switch_on_time = None

        self.is_jetson = host.is_jetson()
        self.is_rpi = host.is_rpi()

        if self.is_rpi:
            import RPi.GPIO as GPIO

            self.switch_pin = 12

            self.green_switch_pin = 16
            self.red_switch_pin = 13

            self.looper_pin = 22  # red led
            self.incoming_pin = 18  # yellow led
            self.data_pin = 36  # green led
            self.outgoing_pin = 40  # blue led

            self.green_led_pin = 15
            self.red_led_pin = 7

            GPIO.setmode(GPIO.BOARD)  # numbers GPIOs by physical location

        if self.is_jetson:
            import Jetson.GPIO as GPIO

            self.switch_pin = 7

            self.trigger_pin = 12
            self.looper_pin = 29  # red led
            self.incoming_pin = 31  # yellow led
            self.data_pin = 32  # green led
            self.outgoing_pin = 33  # blue led

            GPIO.setmode(GPIO.BOARD)  # numbers GPIOs by physical location

        # http://razzpisampler.oreilly.com/ch07.html
        for pin in self.input_pins:
            self.setup(pin, "input", GPIO.PUD_UP)

        for pin in self.output_pins:
            self.setup(pin, "output")
            self.output(pin, False)

    def clock(self):
        super().clock()

        if self.activated(self.switch_pin):
            if self.switch_on_time is None:
                self.switch_on_time = time.time()
                logger.info(f"{NAME}: switch_on_time was set.")
        else:
            self.switch_on_time = None

        if self.switch_on_time is not None:
            self.pulse("outputs")

            if time.time() - self.switch_on_time > 10:
                self.key_buffer += ["s"]

        return self

    def input(self, pin):
        if pin == -1:
            return True

        if self.is_rpi:
            import RPi.GPIO as GPIO
        elif self.is_jetson:
            import Jetson.GPIO as GPIO
        else:
            return False

        return GPIO.input(pin) == GPIO.HIGH

    def output(self, pin, output):
        if pin == -1:
            return self

        if self.is_rpi:
            import RPi.GPIO as GPIO
        elif self.is_jetson:
            import Jetson.GPIO as GPIO
        else:
            return self

        GPIO.output(pin, GPIO.HIGH if output else GPIO.LOW)
        return self

    def release(self):
        if self.is_rpi:
            import RPi.GPIO as GPIO
        elif self.is_jetson:
            import Jetson.GPIO as GPIO
        else:
            return self

        logger.info(f"{self.__class__.__name__}.release()")

        GPIO.cleanup()

        return self

    def setup(self, pin, what, pull_up_down=None):
        if pin == -1:
            return self

        if self.is_rpi:
            import RPi.GPIO as GPIO
        elif self.is_jetson:
            import Jetson.GPIO as GPIO
        else:
            return self

        if what == "output":
            what = GPIO.OUT
        elif what == "input":
            what = GPIO.IN
        else:
            raise NameError(
                f"{self.__class__.__name__}.setup({pin}): unknown what: {what}."
            )

        if pull_up_down is None:
            GPIO.setup(pin, what)
        else:
            GPIO.setup(pin, what, pull_up_down=pull_up_down)

        return self

    def signature(self):
        return super().signature() + (
            [
                "switch:{}".format(
                    string.pretty_duration(
                        time.time() - self.switch_on_time,
                        largest=True,
                        short=True,
                    )
                )
            ]
            if self.switch_on_time is not None
            else []
        )



from typing import Dict, List, Iterator, Tuple
import copy
import os
import numpy as np
import cv2
from functools import reduce
from tqdm import tqdm

from blueness import module
from bluer_options.logger import log_list
from bluer_objects import file
from bluer_objects import README
from bluer_objects.README.consts import assets_path, assets_url
from bluer_objects.logger.image import log_image_grid

from bluer_sbc.host import signature
from bluer_sbc import NAME
from bluer_sbc.parts.classes.part import Part
from bluer_sbc.logger import logger

NAME = module.name(__file__, NAME)


class PartDB:
    def __init__(self):
        self._db: Dict[str, Part] = {}

        suffix = "bluer-sbc/parts"
        self.url_prefix = assets_url(
            suffix=suffix,
            volume=2,
        )
        self.path = assets_path(
            suffix=suffix,
            volume=2,
        )

    def __iter__(self):
        return iter(self._db.values())

    def __setitem__(
        self,
        name: str,
        part: Part,
    ):
        assert isinstance(part, Part)

        self._db[name] = copy.deepcopy(part)
        self._db[name].name = name

    def __getitem__(self, name: str) -> Part:
        return self._db[name]

    def items(self) -> Iterator[Tuple[str, Part]]:
        return self._db.items()

    @property
    def README(self) -> List[str]:
        return sorted(
            [
                "1. [{}](./{}.md).".format(
                    part.info[0],
                    part.name,
                )
                for part in self
            ]
        )

    def adjust(
        self,
        generate_grid: bool = True,
        dryrun: bool = True,
        verbose: bool = False,
    ) -> bool:
        logger.info(
            "{}.adjust{}".format(
                NAME,
                " [dryrun]" if dryrun else "",
            )
        )

        list_of_filenames = reduce(
            lambda x, y: x + y,
            [
                [part.images[0]]
                for part_name, part in self._db.items()
                if part_name != "template" and part.images
            ],
            [],
        )
        log_list(logger, "adjusting", list_of_filenames, "images")

        if generate_grid:
            if not log_image_grid(
                items=[
                    {
                        "filename": os.path.join(self.path, part.images[0]),
                        "title": part_name,
                    }
                    for part_name, part in self._db.items()
                    if part_name != "template" and part.images
                ],
                filename=assets_path(
                    suffix="bluer-sbc/parts/grid.png",
                    volume=2,
                ),
                scale=3,
                header=[
                    "{} part(s)".format(len(self._db) - 1),
                ],
                footer=signature(),
            ):
                return False

        max_width = 0
        max_height = 0
        for filename in tqdm(list_of_filenames):
            success, image = file.load_image(
                os.path.join(self.path, filename),
                log=verbose,
            )
            if not success:
                return success

            max_height = max(max_height, image.shape[0])
            max_width = max(max_width, image.shape[1])

        logger.info(f"size: {max_height} x {max_width}")

        for filename in tqdm(list_of_filenames):
            success, image = file.load_image(
                os.path.join(self.path, filename),
                log=verbose,
            )
            if not success:
                return success

            if image.shape[0] == max_height and image.shape[1] == max_width:
                logger.info("✅")
                continue

            image = image[:, :, :3]

            scale = min(
                max_height / image.shape[0],
                max_width / image.shape[1],
            )
            image = cv2.resize(
                image,
                dsize=(
                    int(scale * image.shape[1]),
                    int(scale * image.shape[0]),
                ),
                interpolation=cv2.INTER_LINEAR,
            )

            padded_image = (
                np.ones(
                    (max_height, max_width, 3),
                    dtype=np.uint8,
                )
                * 255
            )

            y_offset = (max_height - image.shape[0]) // 2
            x_offset = (max_width - image.shape[1]) // 2

            padded_image[
                y_offset : y_offset + image.shape[0],
                x_offset : x_offset + image.shape[1],
            ] = image

            if not dryrun:
                if not file.save_image(
                    os.path.join(self.path, filename),
                    padded_image,
                    log=verbose,
                ):
                    return False

        return True

    def as_images(
        self,
        dict_of_parts: Dict[str, str],
        reference: str = "../../parts",
    ) -> List[str]:
        return README.Items(
            [
                {
                    "name": self._db[part_name].info[0],
                    "marquee": self._db[part_name].image_url(
                        url_prefix=self.url_prefix
                    ),
                    "description": description,
                    "url": f"{reference}/{part_name}.md",
                }
                for part_name, description in dict_of_parts.items()
            ],
            sort=True,
        )

    def as_list(
        self,
        dict_of_parts: Dict[str, str],
        reference: str = "../../parts",
        log: bool = True,
    ) -> List[str]:
        if log:
            logger.info(
                "{}.as_list: {}".format(
                    self.__class__.__name__,
                    ", ".join(dict_of_parts.keys()),
                )
            )

        for part_name in dict_of_parts:
            if part_name not in self._db:
                logger.error(f"{part_name}: part not found.")
                assert False

        return sorted(
            [
                (
                    "1. [{}]({}){}.".format(
                        self._db[part_name].info[0],
                        f"{reference}/{part_name}.md",
                        ": {}".format(description) if description else "",
                    )
                )
                for part_name, description in dict_of_parts.items()
            ]
        )



from bluer_objects import README
from bluer_objects.README.items import ImageItems

from bluer_sbc.README.designs.consts import assets2
from bluer_sbc.designs.swallow.parts import dict_of_parts as parts


image_template = assets2 + "swallow/design/v5/{}?raw=true"

marquee = README.Items(
    [
        {
            "name": "swallow",
            "marquee": image_template.format("01.jpg"),
            "url": "./bluer_sbc/docs/swallow.md",
        }
    ]
)

items = ImageItems(
    {image_template.format(f"{index+1:02}.jpg"): "" for index in range(6)}
)



import argparse

from blueness import module

from bluer_sbc import NAME
from bluer_sbc.session.classes import Session
from bluer_sbc.logger import logger

NAME = module.name(__file__, NAME)


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    help="start",
)
args = parser.parse_args()

success = False
if args.task == "start":
    success = Session.start()
else:
    logger.error(f"-{NAME}: {args.task}: command not found.")

if not success:
    logger.error(f"-{NAME}: {args.task}: failed.")



from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont

from bluer_options import string

from bluer_sbc.hardware.screen import Screen
from bluer_sbc.logger import logger

BUTTON = 24

RST = None  # on the PiOLED this pin isnt used


class Grove(Screen):
    def __init__(self):
        super(Grove, self).__init__()

        from grove.grove_button import GroveButton

        # https://wiki.seeedstudio.com/Grove-OLED_Display_0.96inch/
        self.size = (64, 128)

        self.button = GroveButton(BUTTON)
        self.button.on_press = lambda t: grove_button_on_press(self, t)
        self.button.on_release = lambda t: grove_button_on_release(self, t)

        import Adafruit_SSD1306

        # https://github.com/IcingTomato/Seeed_Python_SSD1315/blob/master/examples/stats.py
        self.display = Adafruit_SSD1306.SSD1306_128_64(rst=RST)

        self.display.begin()
        self.display.clear()
        self.display.display()

        self.image = Image.new(
            "1",
            (self.display.width, self.display.height),
        )

        self.draw = ImageDraw.Draw(self.image)

        # Draw a black filled box to clear the image.
        self.draw.rectangle(
            (0, 0, self.display.width, self.display.height),
            outline=0,
            fill=0,
        )

        self.padding = -2
        self.top = self.padding
        self.bottom = self.display.height - self.padding

        self.font = ImageFont.load_default()

        self.line_count = 8
        self.line_length = 21

    def update_screen(self, image, session, header):
        super().update_screen(image, session, header)

        signature = (" | ".join(session.signature())).split(" | ")

        self.draw.rectangle(
            (0, 0, self.display.width, self.display.height),
            outline=0,
            fill=0,
        )

        for row in range(min(len(signature), self.line_count)):
            self.draw.text(
                (0, self.top + 8 * row),
                signature[row],
                font=self.font,
                fill=255,
            )

        self.display.image(self.image)
        self.display.display()

        return self


def grove_button_on_press(screen: Screen, t):
    logger.info("grove.button: pressed.")


def grove_button_on_release(screen: Screen, t):
    logger.info(f"grove.button: released after {string.pretty_duration(t)}.")

    if t > 60:
        logger.info("long press, ignored.")
        return

    if t > 5:
        key = "s"
    elif t > 3:
        key = "u"
    else:
        key = " "

    screen.key_buffer.append(key)
    logger.info(f"{screen.__class__.__name__}: '{key}'")



from bluer_objects import README
from bluer_objects.README.items import ImageItems

from bluer_sbc.README.designs.consts import assets2

image_template = assets2 + "nafha/{}?raw=true"

marquee = README.Items(
    [
        {
            "name": "nafha",
            "marquee": image_template.format("01.png"),
            "url": "./bluer_sbc/docs/nafha",
        }
    ]
)

items = ImageItems(
    {
        image_template.format(f"{filename}"): ""
        for filename in [
            f"{index+1:02}.png"
            for index in range(
                4,
            )
        ]
        + [
            "20251028_123428.jpg",
            "20251028_123438.jpg",
        ]
    },
)



from bluer_objects.README.items import Items
from bluer_objects import markdown

from bluer_sbc.parts.db import db_of_parts


items = markdown.generate_table(
    Items(
        [
            {
                "name": "parts",
                "url": "./bluer_sbc/docs/parts",
                "marquee": f"{db_of_parts.url_prefix}/grid.png",
            },
        ]
    ),
    log=False,
)



import cv2
from typing import Tuple, List, Union
import numpy as np

from blueness import module
from bluer_options import string
from bluer_options import host
from bluer_options.timer import Timer
from bluer_options.logger import crash_report
from bluer_objects import file, objects

from bluer_sbc import env
from bluer_sbc.hardware import hardware
from bluer_sbc.imager.classes import Imager
from bluer_sbc.logger import logger


class Camera(Imager):
    def __init__(self):
        self.device = None
        self.resolution = []

    def capture(
        self,
        close_after: bool = True,
        log: bool = True,
        open_before: bool = True,
        filename: str = "",
        object_name: str = "",
    ) -> Tuple[bool, np.ndarray]:
        success = False
        image = np.ones((1, 1, 3), dtype=np.uint8) * 127

        if open_before:
            if not self.open():
                return success, image

        if self.device is None:
            return success, image

        success, image = self.capture_function()
        if not success:
            logger.warning("capture failed.")
        elif log:
            logger.info(
                "{}.capture(): {}".format(
                    self.__class__.__name__,
                    string.pretty_shape_of_matrix(image),
                )
            )

        if close_after:
            self.close()

        if success and filename:
            success = file.save_image(
                filename=objects.path_of(
                    object_name=object_name,
                    filename=filename,
                ),
                image=image,
                log=log,
            )

        return success, image

    def capture_function(self) -> Tuple[bool, np.ndarray]:
        success = False
        image = np.ones((1, 1, 3), dtype=np.uint8) * 127

        try:
            success, image = self.device.read()

            if success:
                image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)

        except Exception as e:
            crash_report(e)

        return success, image

    # https://projects.raspberrypi.org/en/projects/getting-started-with-picamera/6
    def capture_video(
        self,
        filename: str,
        object_name: str,
        length: int = 10,  # in seconds
        preview: bool = True,
        pulse: bool = True,
        resolution=None,
    ) -> bool:
        logger.error(f"{self.__class__.__name__}.capture_video(): not implemented.")
        return False

    def close(self, log: bool = True) -> bool:
        if self.device is None:
            logger.warning(
                "{}.close(): device is {}, failed.".format(
                    self.__class__.__name__,
                    self.device,
                )
            )
            return False

        try:
            self.close_function()
        except Exception as e:
            crash_report(e)
            return False

        self.device = None

        if log:
            logger.info(f"{self.__class__.__name__}.close().")

        return True

    def close_function(self):
        self.device.release()

    def get_resolution(self) -> List[int]:
        try:
            resolution = self.get_resolution_function()
        except Exception as e:
            crash_report(e)
            return []

        return resolution

    def get_resolution_function(self) -> Tuple[int, int]:
        return [
            int(self.device.get(const))
            for const in [cv2.CAP_PROP_FRAME_HEIGHT, cv2.CAP_PROP_FRAME_WIDTH]
        ]

    def open(
        self,
        log: bool = True,
        resolution: Union[List[int], None] = None,
    ) -> bool:
        try:
            self.open_function(resolution=resolution)

            self.resolution = self.get_resolution()

            if log:
                logger.info(
                    "{}.open({})".format(
                        self.__class__.__name__,
                        string.pretty_shape(self.resolution),
                    )
                )

            return True
        except Exception as e:
            crash_report(e)
            return False

    def open_function(
        self,
        resolution=Union[List[int], None],
    ):
        self.device = cv2.VideoCapture(0)

        # https://stackoverflow.com/a/31464688
        self.device.set(cv2.CAP_PROP_FRAME_WIDTH, env.BLUER_SBC_CAMERA_WIDTH)
        self.device.set(cv2.CAP_PROP_FRAME_HEIGHT, env.BLUER_SBC_CAMERA_HEIGHT)

    def preview(
        self,
        length: float = -1,
    ) -> bool:
        logger.info(
            "{}.preview{} ... | press q or e to quit ...".format(
                self.__class__.__name__,
                "[{}]".format("" if length == -1 else string.pretty_duration(length)),
            )
        )

        hardware.sign_images = False
        timer = Timer(length, "preview")
        try:
            self.open(
                log=True,
                resolution=(320, 240),
            )

            while not hardware.pressed("qe"):
                _, image = self.capture(
                    close_after=False,
                    log=False,
                    open_before=False,
                )
                hardware.update_screen(image, None, [])

                if timer.tick(wait=True):
                    logger.info(
                        "{} is up, quitting.".format(string.pretty_duration(length))
                    )
                    break

        except KeyboardInterrupt:
            logger.info("Ctrl+C, stopping.")

        finally:
            self.close(log=True)

        return True



from bluer_sbc.parts.db import db_of_parts

docs = [
    {
        "path": "../docs/parts",
        "macros": {"list:::": db_of_parts.README},
    }
] + [
    {
        "path": part.filename(create=True),
        "macros": {
            "info:::": part.README(db_of_parts.url_prefix),
        },
    }
    for part_name, part in db_of_parts.items()
    if part_name != "template"
]



from typing import List

from bluer_options.terminal import show_usage, xtra

from bluer_sbc import ALIAS


def help_cd(
    tokens: List[str],
    mono: bool,
) -> str:
    return show_usage(
        [
            "@sbc",
            "parts",
            "cd",
        ],
        "cd to part images.",
        mono=mono,
    )


def help_adjust(
    tokens: List[str],
    mono: bool,
) -> str:
    options = xtra("dryrun,~grid", mono=mono)

    args = [
        "[--verbose 1]",
    ]

    return show_usage(
        [
            "@sbc",
            "parts",
            "adjust",
            f"[{options}]",
        ]
        + args,
        "adjust part images.",
        mono=mono,
    )


help_functions = {
    "adjust": help_adjust,
    "cd": help_cd,
}



from bluer_sbc.README.design import design
from bluer_sbc.README.designs import (
    cheshmak,
    battery_bus,
    adapter_bus,
    swallow,
    swallow_head,
    bryce,
    nafha,
    shelter,
    ultrasonic_sensor_tester,
    x,
)


docs = [
    {
        "items": design_info["items"],
        "path": f"../docs/{design_name}.md",
        "macros": design_info.get("macros", {}),
    }
    for design_name, design_info in {
        "battery-bus": design(
            battery_bus.items,
            battery_bus.parts,
        ),
        "adapter-bus": design(
            adapter_bus.items,
            adapter_bus.parts,
        ),
        "bryce": design(
            bryce.items,
        ),
        "cheshmak": design(
            cheshmak.items,
        ),
        "shelter": design(
            shelter.items,
        ),
        "swallow-head": design(
            swallow_head.items,
            swallow_head.parts,
        ),
        "swallow": design(
            swallow.items,
            swallow.parts,
        ),
        "ultrasonic-sensor-tester": design(
            ultrasonic_sensor_tester.items,
        ),
        "x": design(
            x.items,
        ),
    }.items()
] + [
    {
        "items": nafha.items,
        "path": "../docs/nafha",
    },
    {
        "path": "../docs/nafha/parts-v1.md",
    },
]



from typing import List, Dict

from bluer_objects import markdown

from bluer_sbc.parts.db import db_of_parts


def design(
    items: List[str],
    dict_of_parts: Dict = {},
    macros: Dict = {},
) -> Dict:
    output = {
        "items": items,
    }

    if dict_of_parts:
        output["macros"] = {
            "parts_images:::": markdown.generate_table(
                db_of_parts.as_images(
                    dict_of_parts,
                    reference="./parts",
                ),
                cols=10,
                log=False,
            ),
            "parts_list:::": db_of_parts.as_list(
                dict_of_parts,
                reference="./parts",
                log=False,
            ),
        }

        output["macros"].update(macros)

    return output



from bluer_objects import README
from bluer_objects.README.items import ImageItems
from bluer_objects.README.consts import assets_url


assets2_x = assets_url(
    suffix="x",
    volume=2,
)

marquee = README.Items(
    [
        {
            "name": "x",
            "marquee": f"{assets2_x}/TBA.jpg",
            "url": "./bluer_sbc/docs/x.md",
        }
    ]
)

items = ImageItems(
    {
        f"{assets2_x}/TBA.jpg": "",
        f"{assets2_x}/TBA.jpg": "",
    }
)



import argparse
import time

from blueness import module
from bluer_options import string

from bluer_sbc import NAME
from bluer_sbc.hardware.hat.prototype import Prototype_Hat
from bluer_sbc.logger import logger

NAME = module.name(__file__, NAME)


parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    default="",
    help="input/output",
)
parser.add_argument(
    "--outputs",
    type=str,
    default="",
)
args = parser.parse_args()

hardware = Prototype_Hat()

success = False
if args.task == "input":
    logger.info("loop started (Ctrl+C to stop)")
    # https://stackoverflow.com/a/18994932/10917551
    try:
        while True:
            logger.info(
                "inputs: {}".format(
                    ", ".join([str(hardware.input(pin)) for pin in hardware.input_pins])
                )
            )
            time.sleep(0.1)
    except KeyboardInterrupt:
        logger.info("Ctrl+C, stopping.")
    finally:
        hardware.release()
    success = True
elif args.task == "output":
    outputs = args.outputs + len(hardware.output_pins) * "1"
    for index, pin in enumerate(hardware.output_pins):
        hardware.output(pin, outputs[index] == "1")
    hardware.release()
    success = True
elif args.task == "validate":
    logger.info("loop started (Ctrl+C to stop)")
    value = True
    try:
        while True:
            activity = False
            for pin, pin_name in zip(
                [
                    hardware.green_switch_pin,
                    hardware.red_switch_pin,
                    hardware.switch_pin,
                    hardware.trigger_pin,
                ],
                "green_switch_pin,red_switch_pin,switch_pin,trigger_pin".split(","),
            ):
                if hardware.activated(pin):
                    logger.info(
                        "{}: {} activated.".format(
                            string.timestamp(ms=True),
                            pin_name,
                        )
                    )
                    activity = True

            for pin in hardware.output_pins:
                hardware.output(pin, activity or value)
            time.sleep(0.1)

            value = not value
    except KeyboardInterrupt:
        logger.info("Ctrl+C, stopping.")
    finally:
        hardware.release()
    success = True
else:
    logger.error(f"-{NAME}: {args.task}: command not found.")

if not success:
    logger.error(f"-{NAME}: {args.task}: failed.")



import argparse
import os

from blueness import module

from bluer_sbc import NAME
from bluer_sbc.imager.lepton import instance as lepton
from bluer_sbc.hardware import hardware
from bluer_sbc.logger import logger

NAME = module.name(__file__, NAME)

parser = argparse.ArgumentParser(NAME)
parser.add_argument(
    "task",
    type=str,
    default="",
    help="capture|preview",
)
parser.add_argument(
    "--filename",
    default="",
    type=str,
)
parser.add_argument(
    "--output_path",
    type=str,
    default="",
)
args = parser.parse_args()

success = False
if args.task == "capture":
    success, _, _ = lepton.capture(
        filename=os.path.join(args.output_path, "camera.jpg"),
    )
elif args.task == "preview":
    success = True

    hardware.sign_images = False
    try:
        while not hardware.pressed("qe"):
            _, image = lepton.capture()
            hardware.update_screen(image)
    finally:
        pass
else:
    logger.error(f"-{NAME}: {args.task}: command not found.")

if not success:
    logger.error(f"-{NAME}: {args.task}: failed.")



from bluer_objects import README
from bluer_objects.README.items import ImageItems
from bluer_objects.README.consts import assets_url
from bluer_objects.README.consts import designs_url

from bluer_sbc.designs.battery_bus.parts import dict_of_parts as parts

assets2 = assets_url(
    suffix="battery-bus",
    volume=2,
)

marquee = README.Items(
    [
        {
            "name": "battery bus",
            "marquee": f"{assets2}/20251007_221902.jpg",
            "url": "./bluer_sbc/docs/battery-bus.md",
        }
    ]
)

items = ImageItems(
    {
        f"{assets2}/concept.png": "",
        designs_url(
            "battery-bus/electrical/wiring.png?raw=true",
        ): designs_url(
            "battery-bus/electrical/wiring.svg",
        ),
        f"{assets2}/20251007_221902.jpg": "",
        f"{assets2}/20251007_220642.jpg": "",
        f"{assets2}/20251007_220520.jpg": "",
        f"{assets2}/20251007_220601.jpg": "",
    }
)



from bluer_objects import README
from bluer_objects.README.items import ImageItems

from bluer_sbc.README.designs.consts import assets2

image_template = assets2 + "cheshmak/{}?raw=true"

marquee = README.Items(
    [
        {
            "name": "cheshmak",
            "marquee": image_template.format("01.png"),
            "url": "./bluer_sbc/docs/cheshmak.md",
        }
    ]
)

items = ImageItems(
    {image_template.format(f"{index+1:02}.png"): "" for index in range(1)}
)



from typing import List, Union
import copy

from bluer_objects import markdown
from bluer_objects import file


class Part:
    def __init__(
        self,
        info: Union[List[str], str] = [],
        name: str = "",
        images: List[str] = [],
    ):
        self.name = name

        self.info = (
            copy.deepcopy(info)
            if isinstance(
                info,
                list,
            )
            else [info]
        )

        self.images = (
            copy.deepcopy(images)
            if isinstance(
                images,
                list,
            )
            else [images]
        )

    def filename(
        self,
        create: bool = False,
    ) -> str:
        reference = file.path(__file__)
        filename = file.absolute(
            f"../../docs/parts/{self.name}.md",
            reference,
        )

        if not create or file.exists(filename):
            return filename

        template_filename = file.absolute(
            "../../docs/parts/template-template.md",
            reference,
        )
        assert file.copy(
            template_filename,
            file.add_suffix(filename, "template"),
            log=True,
        ), template_filename

        return filename

    def image_url(
        self,
        url_prefix: str,
        filename: str = "",
    ) -> str:

        return (
            "{}/{}?raw=true".format(
                url_prefix,
                filename if filename else self.images[0],
            )
            if self.images
            else ""
        )

    def README(
        self,
        url_prefix: str,
    ) -> List[str]:
        return [f"- {info}" for info in self.info] + (
            [""]
            + markdown.generate_table(
                [
                    "![image]({})".format(
                        self.image_url(
                            url_prefix,
                            filename,
                        )
                    )
                    for filename in self.images
                ],
                cols=3,
                log=False,
            )
            if self.images
            else []
        )
